"""Mann-Whitney-related tests.
"""

import scipy.stats as s

import statys.utils.wrappers as w
from statys.utils import logging

logger = logging.get_logger(__name__)


def u_test(dist, alpha=0.05):
    """Performs the Mann-Whitney U test.

    Args:
        dist (Distribution): Distribution to be analyzed.
        alpha (float): Significance value.

    Returns:
        Dictionary holding the test's outputs.

    """

    logger.info("Performing Mann-Whitney U test ...")

    output = w.statistical_pipeline(s.mannwhitneyu, dist, alpha)

    logger.info("Test performed.")
    logger.debug(output)

    return output



"""Critical difference-based plotting utilities. Note that is an improved version of the script provided
by https://github.com/biolab/orange3.
"""


from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from matplotlib.axis import Axis
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure


def _create_labels(size: Optional[int] = 1) -> List[str]:
    """Creates a list of labels strings.

    Args:
        size: Amount of strings to be created.

    Returns:
        (List[str]): Stringed labels, e.g., x0, x1, ..., xn.

    """

    labels = []

    for i in range(size):
        labels.append(f"$x_{{{i}}}$")

    return labels


def _multiple_range(input_list: List[Any]) -> Tuple[Any, ...]:
    """Performs a multiple range, used to traverse matrices.

    Args:
        input_list: List to be traversed.

    Yields:
        (Tuple[Any, ...]): Iterator of tuples.

    """

    length = len(input_list)

    if length == 0:
        yield ()

    else:
        index = input_list[0]

        if isinstance(index, int):
            index = [index]

        for a in range(*index):
            for b in _multiple_range(input_list[1:]):
                yield tuple([a] + list(b))


def _line_factoring(line: List[float], factor: float) -> List[float]:
    """Factors a line's positioning.

    Args:
        line: Lines to be factored.
        factor: Factor to use in the factoring.

    Returns:
        (List[float]): Factored lines.

    """

    return [a * factor for a in line]


def _plot_line(
    ax: Axis,
    input_list: List[Tuple[int, int]],
    width_factor: float,
    height_factor: float,
    color: Optional[str] = "k",
    **kwargs,
) -> None:
    """Plots pairs of points as lines.

    Args:
        ax: Axis to be plotted.
        input_list: List of pairs of points.
        width_factor: Width factor.
        height_factor: Height factor.
        color: Color to be plotted.

    """

    # Calculates the `x` and `y` elements
    x = _line_factoring(_get_element(input_list, 0), width_factor)
    y = _line_factoring(_get_element(input_list, 1), height_factor)

    # Plots the pair of points
    ax.plot(x, y, color=color, **kwargs)


def _plot_text(
    ax: Axis,
    x: int,
    y: int,
    s: str,
    width_factor: float,
    height_factor: float,
    **kwargs,
) -> None:
    """Plots a text on the desired position.

    Args:
        ax: Axis to be plotted.
        x: `x` position to be plotted.
        y: `y` position to be plotted.
        s: String to be plotted.
        width_factor: Width factor.
        height_factor: Height factor.

    """

    # Plots the text itself
    ax.text(width_factor * x, height_factor * y, s, **kwargs)


def _get_amount_lines(ranks: List[int], cd: float) -> List[Tuple[Any, ...]]:
    """Gets the amount of possible lines to create the plot.

    Args:
        ranks: List of ranks.
        cd: Critical difference.

    Returns:
        (List[Tuple[Any, ...]]): Longest possible amount of lines.

    """

    # Calculates the number of possible ranks
    n_ranks = len(ranks)

    # Transforms them into pairs
    pairs = [(i, j) for i, j in _multiple_range([[n_ranks], [n_ranks]]) if j > i]

    # Remove non-significant pairs
    non_pairs = [(i, j) for i, j in pairs if abs(ranks[i] - ranks[j]) <= cd]

    # Internal function to return the longest pairs
    def get_longest(i: int, j: int, pairs: Tuple[int, int]) -> bool:
        for k, l in pairs:
            # Checks if iterated pair is bigger than current one
            if (k <= i and l > j) or (k < i and l >= j):
                return False

        return True

    # Keeps the longest pair
    longest = [(i, j) for i, j in non_pairs if get_longest(i, j, non_pairs)]

    return longest


def _get_element(input_list: List[Any], n: int) -> List[Any]:
    """Gets a specific element from a list of tuples.

    Args:
        input_list: List to be iterated.
        n: Element to be retrieved.

    Returns:
        (List[Any]): Only retrieved elements.

    """

    if n < 0:
        # Adds the length of the list to gather the desired index
        i = len(input_list[0]) + n

    else:
        # Use the index as usual
        i = n

    return [a[i] for a in input_list]


def _calculate_plot_properties(
    sort_ranks: List[int], cd: float
) -> Tuple[int, float, float, int, float, float]:
    """Calculates a set of properties used to accurately plot the statistical test.

    Args:
        sort_ranks: List holding the sorted ranks.
        cd: Critical difference.

    Returns:
        (Tuple[int, float, float, int, float, float]): Properties used in the plot's construction.

    """

    # Gathers the number of possible ranks and (longest) lines
    n_ranks = len(sort_ranks)
    n_lines = _get_amount_lines(sort_ranks, cd)

    # Instantiates the height distance as 0.25 and calculates top distance
    height_distance = 0.25
    top_distance = height_distance + 0.4

    # Calculates the value of the blanked lines
    blank_lines = 0.2 + 0.2 + (len(n_lines) - 1) * 0.1

    # Gathers the position of non-significant samples
    not_sig_distance = max(2 * 0.2, blank_lines)

    # Finally, calculates the plot's height
    height = top_distance + ((n_ranks + 1) / 2) * 0.2 + not_sig_distance

    return n_ranks, height_distance, top_distance, n_lines, not_sig_distance, height


def _prepare_plot(width: float, height: float) -> Tuple[Figure, Axis]:
    """Prepares the plot prior to its use.

    Args:
        width: Width of the plot.
        height: Height of the plot.

    Returns:
        (Tuple[Figure, Axis]): Figure and axis properties from the plot.

    """

    # Creates a new figure based on inputted width and height
    fig = Figure(figsize=(width, height))

    # Adds the corresponding axes to create a box
    ax = fig.add_axes([0, 0, 1, 1])

    # Sets the outer box display as off
    ax.set_axis_off()

    # Plots its boundaries
    ax.plot([0, 1], [0, 1], c="w")

    # Sets the `x` and `y` limits
    ax.set_xlim(0, 1)
    ax.set_ylim(1, 0)

    return fig, ax


def _position_rank(
    index: int, low: int, high: int, text_spacing: int, scale: float, reverse: bool
) -> int:
    """Positions a rank according to its value.

    Args:
        index: Index to be positioned.
        low: Minimum rank possible.
        high: Maximum rank possible.
        text_spacing: Text spacing inside the plot.
        scale: Plot's scale.
        reverse: Whether plot should use ascending or descending order.

    Returns:
        (int): Rank should be positioned in the plot.

    """

    if reverse:
        x = high - index

    else:
        x = index - low

    return text_spacing + scale / (high - low) * x


def plot_critical_difference(
    cd_dict: Dict[str, Any],
    labels: Optional[List[str]] = None,
    width: Optional[int] = 6,
    text_spacing: Optional[int] = 2,
    reverse: Optional[bool] = False,
) -> None:
    """Plots the critical difference between the averaged ranks.

    Args:
        cd_dict: Dictionary of average ranks and critical differences.
        labels: List of stringed labels.
        width: Plot's width.
        text_spacing: Text spacing inside the plot.
        reverse: Whether plot should use ascending or descending order.

    """

    # Defines the plot's scale
    scale = width - 2 * text_spacing

    for key, v in cd_dict.items():
        # Gathers the minimum and maximum possible ranks
        low, high = 1, v[0].shape[0]

        # Gathers current item ranks and its critical difference
        ranks, cd = v[0], v[1]

        # Defines a sorting-based operator
        sort_operator = sorted([(r, i) for i, r in enumerate(ranks)], reverse=reverse)

        # Gathers the sorted ranks and their indexes
        sort_ranks, sort_idx = _get_element(sort_operator, 0), _get_element(
            sort_operator, 1
        )

        # Checks if labels exists and number of supplied labels equals the number of arguments
        if labels and len(labels) == len(ranks):
            pass

        else:
            # Re-creates the labels list
            labels = _create_labels(len(ranks))

        # Gathers the sorted labels as well
        sort_labels = [labels[i] for i in sort_idx]

        # Calculates a set of properties used to perform an accurate plot
        (
            n_ranks,
            height_distance,
            top_distance,
            n_lines,
            not_sig_distance,
            height,
        ) = _calculate_plot_properties(sort_ranks, cd)

        # Prepares the plot based on inputted width and calculated height
        fig, ax = _prepare_plot(width, height)

        # Calculates the width and height factors
        width_factor = 1 / width
        height_factor = 1 / height

        # Defines `x` and `y` points
        x = (text_spacing, top_distance)
        y = (width - text_spacing, top_distance)

        # Plots the first line
        _plot_line(ax, [x, y], width_factor, height_factor, linewidth=0.7)

        # Defines regular, big and small ticks sizes
        big_tick = 0.1
        small_tick = 0.05

        # Iterates over every possible value between low and high
        # for plotting the ticks
        for a in list(np.arange(low, high, 0.5)) + [high]:
            # Gathers the current tick as smallest one
            tick = small_tick

            if isinstance(a, int):
                # Gathers the current tick as biggest one
                tick = big_tick

            # Defines `x` and `y` points
            x = (
                _position_rank(a, low, high, text_spacing, scale, reverse),
                top_distance - tick / 2,
            )
            y = (
                _position_rank(a, low, high, text_spacing, scale, reverse),
                top_distance,
            )

            # Plots a new line
            _plot_line(ax, [x, y], width_factor, height_factor, linewidth=0.7)

        # Iterates over every possible value between low and high
        # for plotting the text
        for a in range(low, high + 1):
            # Plots the text label
            _plot_text(
                ax,
                _position_rank(a, low, high, text_spacing, scale, reverse),
                top_distance - tick / 2 - 0.05,
                str(a),
                width_factor,
                height_factor,
                ha="center",
                va="bottom",
            )

        # Iterates over every possible left-sided rank
        for i in range(int((n_ranks + 1) / 2)):
            # Calculates the "line-arrow"
            arrow = top_distance + not_sig_distance + i * 0.2

            # Defines `x`, `y` and `z` points
            x = (
                _position_rank(sort_ranks[i], low, high, text_spacing, scale, reverse),
                top_distance,
            )
            y = (
                _position_rank(sort_ranks[i], low, high, text_spacing, scale, reverse),
                arrow,
            )
            z = (text_spacing - 0.1, arrow)

            # Plots a new line
            _plot_line(ax, [x, y, z], width_factor, height_factor, linewidth=0.7)

            # Plots the text label
            _plot_text(
                ax,
                text_spacing - 0.2,
                arrow,
                sort_labels[i],
                width_factor,
                height_factor,
                ha="right",
                va="center",
            )

        # Iterates over every possible right-sided rank
        for i in range(int((n_ranks + 1) / 2), n_ranks):
            # Calculates the "line-arrow"
            arrow = top_distance + not_sig_distance + (n_ranks - i - 1) * 0.2

            # Defines `x`, `y` and `z` points
            x = (
                _position_rank(sort_ranks[i], low, high, text_spacing, scale, reverse),
                top_distance,
            )
            y = (
                _position_rank(sort_ranks[i], low, high, text_spacing, scale, reverse),
                arrow,
            )
            z = (text_spacing + scale + 0.1, arrow)

            # Plots a new line
            _plot_line(ax, [x, y, z], width_factor, height_factor, linewidth=0.7)

            # Plots the text label
            _plot_text(
                ax,
                text_spacing + scale + 0.2,
                arrow,
                sort_labels[i],
                width_factor,
                height_factor,
                ha="left",
                va="center",
            )

        if reverse:
            # Calculates the starting and ending position from `high` values
            start = _position_rank(high, low, high, text_spacing, scale, reverse)
            end = _position_rank(high - cd, low, high, text_spacing, scale, reverse)

        else:
            # Calculates the starting and ending position from `low` values
            start = _position_rank(low, low, high, text_spacing, scale, reverse)
            end = _position_rank(low + cd, low, high, text_spacing, scale, reverse)

        # Plots the starting and ending points of the CD line
        _plot_line(
            ax,
            [(start, height_distance), (end, height_distance)],
            width_factor,
            height_factor,
            linewidth=0.7,
        )

        # Plots the starting ticks of the CD line
        _plot_line(
            ax,
            [
                (start, height_distance + big_tick / 2),
                (start, height_distance - big_tick / 2),
            ],
            width_factor,
            height_factor,
            linewidth=0.7,
        )

        # Plots the ending ticks of the CD line
        _plot_line(
            ax,
            [
                (end, height_distance + big_tick / 2),
                (end, height_distance - big_tick / 2),
            ],
            width_factor,
            height_factor,
            linewidth=0.7,
        )

        # Plots the CD line itself
        _plot_text(
            ax,
            (start + end) / 2,
            height_distance - 0.05,
            f"CD={cd}",
            width_factor,
            height_factor,
            ha="center",
            va="bottom",
        )

        # Plots non-significant lines
        # Defines a new starting point
        start = top_distance + 0.2

        # Iterates over every non-significant line
        for left, right in n_lines:
            # Defines `x` and `y` points
            x = (
                _position_rank(
                    sort_ranks[left], low, high, text_spacing, scale, reverse
                )
                - 0.05,
                start,
            )
            y = (
                _position_rank(
                    sort_ranks[right], low, high, text_spacing, scale, reverse
                )
                + 0.05,
                start,
            )

            # Plots the line
            _plot_line(ax, [x, y], width_factor, height_factor, linewidth=2.5)

            # Adds height to distinguish between lines
            start += 0.1

        # Adding a figure to the canvas
        canvas = FigureCanvasAgg(fig)

        # Printing out the figure
        canvas.print_figure(f"cd_{key}.pdf")



"""Logging-based methods and helpers.
"""

import logging
import sys
from logging import Logger, StreamHandler
from logging.handlers import TimedRotatingFileHandler

FORMATTER = logging.Formatter("%(asctime)s - %(name)s — %(levelname)s — %(message)s")
LOG_FILE = "statys.log"


def get_console_handler() -> StreamHandler:
    """Gets a console handler to handle logging into console.

    Returns:
        (StreamHandler): Handler to output information into console.

    """

    # Creates a stream handler for logger
    console_handler = StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)

    return console_handler


def get_timed_file_handler() -> TimedRotatingFileHandler:
    """Gets a timed file handler to handle logging into files.

    Returns:
        (TimedRotatingFileHandler): Handler to output information into timed files.

    """

    # Creates a timed rotating file handler for logger
    file_handler = TimedRotatingFileHandler(LOG_FILE, delay=True, when="midnight")
    file_handler.setFormatter(FORMATTER)

    return file_handler


def get_logger(logger_name: str) -> Logger:
    """Gets a logger and make it avaliable for further use.

    Args:
        logger_name: The name of the logger.

    Returns:
        (Logger): Logger instance.

    """

    # Creates a logger object
    logger = logging.getLogger(logger_name)

    # Sets an log level
    logger.setLevel(logging.DEBUG)

    # Adds the desired handlers
    logger.addHandler(get_console_handler())
    logger.addHandler(get_timed_file_handler())

    # True or False for propagating logs
    logger.propagate = False

    return logger



"""Significance-based plotting utilities, such as h-index and p-value.
"""

from typing import Any, Dict, List, Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axis import Axis


def _create_labels(size: Optional[int] = 1) -> List[str]:
    """Creates a list of labels strings.

    Args:
        size: Amount of strings to be created.

    Returns:
        (List[str]): Stringed labels, e.g., arg0, arg1, ..., argn.

    """

    labels = []

    for i in range(size):
        labels.append(f"$arg_{{{i}}}$")

    return labels


def _prepare_plot(n_args: int, labels: List[str], title: str) -> Axis:
    """Prepares the plot with common definitions.

    Args:
        n_args: Number of arguments.
        labels: List of stringed labels.
        title: Title to be displayed.

    Returns:
        (Axis): Axis property from the plot.

    """

    # Creates the figure and its axis
    _, ax = plt.subplots()

    # Checks if labels exists and number of supplied labels equals the number of arguments
    if labels and len(labels) == n_args:
        pass

    else:
        # Re-creates the labels list
        labels = _create_labels(n_args)

    # Defines axis properties
    ax.set_xticks(np.arange(0, n_args, 1))
    ax.set_xticklabels(labels)
    ax.set_yticks(np.arange(0, n_args, 1))
    ax.set_yticklabels(labels)
    ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)

    # Defines grid properties
    ax.set_title(title)
    ax.set_xticks(np.arange(n_args + 1) - 0.5, minor=True)
    ax.set_yticks(np.arange(n_args + 1) - 0.5, minor=True)
    ax.grid(which="minor", color="w", linestyle="-", linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)

    # Iterates through every spine
    for _, spine in ax.spines.items():
        # Removes the outline border
        spine.set_visible(False)

    return ax


def plot_p_value(
    p_dict: Dict[str, Any],
    color_map: Optional[str] = "YlOrRd",
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
) -> None:
    """Plots a p-value grid according to statistical results.

    Args:
        p_dict: Significances and p-values.
        color_map: Color map from matplotlib.
        labels: Stringed labels.
        title: Title to be displayed.

    """

    # Calculates the number of arguments by solving: y = x^2 - x
    n_args = round(np.roots([1, -1, -len(p_dict)])[0])

    # Prepares the plot using common-based definitions
    ax = _prepare_plot(n_args, labels, title)

    # Instantiates the p-valued matrix
    p = np.zeros((n_args, n_args))

    for k, v in p_dict.items():
        # Gathers the positions from the arguments
        args = k.replace("arg", "").split("-")

        # Transforms the positions into integers
        i, j = int(args[0]), int(args[1])

        # Replaces its value
        p[i][j] = 1 - v[1]

    # Iterates through the p-valued matrix
    for (i, j), z in np.ndenumerate(p):
        # Applies the corresponding value to the position
        ax.text(j, i, "{:0.3f}".format(1 - z), ha="center", va="center")

    # Adds the significances to the plot
    ax.imshow(p, cmap=color_map)

    # Displays the plot
    plt.show()


def plot_h_index(
    h_dict: Dict[str, Any],
    color_map: Optional[str] = "YlOrRd",
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
) -> None:
    """Plots an h-index grid according to statistical results.

    Args:
        h_dict: H-indexes and p-values.
        color_map: Color map from matplotlib.
        labels: Stringed labels.
        title: Title to be displayed.

    """

    # Calculates the number of arguments by solving: y = x^2 - x
    n_args = round(np.roots([1, -1, -len(h_dict)])[0])

    # Prepares the plot using common-based definitions
    ax = _prepare_plot(n_args, labels, title)

    # Instantiates the significance matrix
    sigs = np.zeros((n_args, n_args), dtype="int")

    for k, v in h_dict.items():
        # Gathers the positions from the arguments
        args = k.replace("arg", "").split("-")

        # Transforms the positions into integers
        i, j = int(args[0]), int(args[1])

        # Replaces its value
        sigs[i][j] = v[0]

    # Iterates through the significance matrix
    for (i, j), z in np.ndenumerate(sigs):
        # Applies the corresponding value to the position
        ax.text(j, i, z, ha="center", va="center")

    # Adds the significances to the plot
    ax.imshow(sigs, cmap=color_map)

    # Displays the plot
    plt.show()



"""Wraps common-based functions for easier development.
"""


from typing import Any, Dict

from statys.core.distribution import Distribution


def calculate_hypothesis(p: float, alpha: float) -> bool:
    """Calculates whether hypothesis' rejection fails or not according to a significance value.

    Args:
        p: P-value from a statistical test.
        alpha: Significance value.

    Returns:
        (bool): 1 if hypothesis is rejected and 0 if it failed to be rejected.

    """

    if p >= alpha:
        # If yes, indicates a failure to reject the null hypothesis
        h = 0

    else:
        # Hypothesis should be rejected
        h = 1

    return h


def measure_pipeline(measure: callable, dist: Distribution, **kwargs) -> Dict[str, Any]:
    """Wraps the pipeline of conducting a measure.

    Args:
        measure: Pointer to a measure function.
        dist: Distribution to be analyzed.

    Returns:
        (Dict[str, Any]): Test's outputs.

    """

    # Initializes a empty dictionary for the outputs
    output = {}

    for (attr, value) in dist.attrs:
        # Calculates the measure
        output[attr] = measure(value, **kwargs)

    return output


def statistical_pipeline(
    test: callable, dist: Distribution, alpha: float
) -> Dict[str, Any]:
    """Wraps the pipeline of conducting a statistical test and calculating its hypothesis.

    Args:
        test: Pointer to a statistical test.
        dist: Distribution to be analyzed.
        alpha: Significance value.

    Returns:
        (Dict[str, Any]): Test's outputs.

    """

    # Initializes a empty dictionary for the outputs
    output = {}

    for (attr, value) in dist.attrs:
        for (attr2, value2) in dist.attrs:
            if attr == attr2:
                pass

            else:
                # Creates a key to the dictionary
                key = attr + "-" + attr2

                # Performs the statistical test
                _, p = test(value, value2)

                # Calculates whether hypothesis fails to be rejected or not
                h = calculate_hypothesis(p, alpha)

                # Adds the tuple to the output dictionary
                output[key] = (h, p)

    return output



"""Statistical-related measures.
"""

import numpy as np
import scipy.stats as s

import statys.utils.wrappers as w
from statys.utils import logging

logger = logging.get_logger(__name__)


def kurtosis(dist, **kwargs):
    """Measures the kurtosis of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Calculating kurtosis ...")

    output = w.measure_pipeline(s.kurtosis, dist, **kwargs)

    logger.info("Kurtosis calculated.")
    logger.debug(output)

    return output


def max(dist, **kwargs):
    """Measures the maximum value of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Finding maximum value ...")

    output = w.measure_pipeline(np.max, dist, **kwargs)

    logger.info("Maximum value found.")
    logger.debug(output)

    return output


def mean(dist, **kwargs):
    """Measures the mean of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Calculating mean ...")

    output = w.measure_pipeline(np.mean, dist, **kwargs)

    logger.info("Mean calculated.")
    logger.debug(output)

    return output


def median(dist, **kwargs):
    """Measures the median of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Calculating median ...")

    output = w.measure_pipeline(np.median, dist, **kwargs)

    logger.info("Median calculated.")
    logger.debug(output)

    return output


def min(dist, **kwargs):
    """Measures the minimum value of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Finding minimum value ...")

    output = w.measure_pipeline(np.min, dist, **kwargs)

    logger.info("Minimum value found.")
    logger.debug(output)

    return output


def rank(dist, **kwargs):
    """Ranks the values of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Ranking distribution ...")

    output = w.measure_pipeline(s.rankdata, dist, **kwargs)

    logger.info("Distribution ranked.")
    logger.debug(output)

    return output


def skewness(dist, **kwargs):
    """Measures the skewness of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Calculating skewness ...")

    output = w.measure_pipeline(s.skew, dist, **kwargs)

    logger.info("Skewness calculated.")
    logger.debug(output)

    return output


def std(dist, **kwargs):
    """Measures the standard deviation of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Calculating standard deviation ...")

    output = w.measure_pipeline(np.std, dist, **kwargs)

    logger.info("Standard deviation calculated.")
    logger.debug(output)

    return output


def var(dist, **kwargs):
    """Measures the variance of a distribution.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the measure's outputs.

    """

    logger.info("Calculating variance ...")

    output = w.measure_pipeline(np.var, dist, **kwargs)

    logger.info("Variance calculated.")
    logger.debug(output)

    return output



"""Distribution-related definitions.
"""

from typing import Any, Dict

import numpy as np

import statys.utils.exception as e
from statys.utils import logging

logger = logging.get_logger(__name__)


class Distribution:
    """A class that serves as the foundation for calculating statistical analysis. In other words,
    one can interpret this class as the population.

    """

    def __init__(self, *args) -> None:
        """Initialization method."""

        logger.info("Initializing class with %d arguments ...", len(args))

        for i, arg in enumerate(args):
            attr = f"arg{i}"

            if not isinstance(arg, (list, np.ndarray)):
                raise e.TypeError(f"`{attr}` should be a list or np.ndarray")

            setattr(self, attr, arg)

        logger.debug("%s", self)
        logger.info("Class initialized.")

    def __repr__(self) -> str:
        """Class' string representation.

        Returns:
            (str): String representation.

        """

        return str(self.__dict__)

    @property
    def attrs(self) -> Dict[str, Any]:
        """Gathers all attributes from class.

        Returns:
            (Dict[str, Any]): Attributes encoded into a dictionary.

        """

        return self.__dict__.items()



"""Wilcoxon-related tests.
"""

import scipy.stats as s

import statys.utils.wrappers as w
from statys.utils import logging

logger = logging.get_logger(__name__)


def signed_rank(dist, alpha=0.05):
    """Performs the Wilcoxon signed-rank test.

    Args:
        dist (Distribution): Distribution to be analyzed.
        alpha (float): Significance value.

    Returns:
        Dictionary holding the test's outputs.

    """

    logger.info("Performing Wilcoxon signed-rank test ...")

    output = w.statistical_pipeline(s.wilcoxon, dist, alpha)

    logger.info("Test performed.")
    logger.debug(output)

    return output


def rank_sum(dist, alpha=0.05):
    """Performs the Wilcoxon rank-sum test.

    Args:
        dist (Distribution): Distribution to be analyzed.
        alpha (float): Significance value.

    Returns:
        Dictionary holding the test's outputs.

    """

    logger.info("Performing Wilcoxon rank-sum test ...")

    output = w.statistical_pipeline(s.ranksums, dist, alpha)

    logger.info("Test performed.")
    logger.debug(output)

    return output



"""Exceptions.
"""

from statys.utils import logging

logger = logging.get_logger(__name__)


class Error(Exception):
    """A generic Error class derived from Exception.

    Essentially, it gets the class and message and logs the error to the logger.

    """

    def __init__(self, cls: str, msg: str) -> None:
        """Initialization method.

        Args:
            cls: Class identifier.
            msg: Message to be logged.

        """

        super(Error, self).__init__()

        # Logs the error in a formatted way
        logger.error("%s: %s.", cls, msg)


class ArgumentError(Error):
    """An ArgumentError class for logging errors related to wrong number of provided arguments."""

    def __init__(self, error: str) -> None:
        """Initialization method.

        Args:
            error: Error message to be logged.

        """

        super(ArgumentError, self).__init__("ArgumentError", error)


class SizeError(Error):
    """A SizeError class for logging errors related to wrong length or size of variables."""

    def __init__(self, error: str) -> None:
        """Initialization method.

        Args:
            error: Error message to be logged.

        """

        super(SizeError, self).__init__("SizeError", error)


class TypeError(Error):
    """A TypeError class for logging errors related to wrong type of variables."""

    def __init__(self, error: str) -> None:
        """Initialization method.

        Args:
            error: Error message to be logged.

        """

        super(TypeError, self).__init__("TypeError", error)


class ValueError(Error):
    """A ValueError class for logging errors related to wrong value of variables."""

    def __init__(self, error: str) -> None:
        """Initialization method.

        Args:
            error: Error message to be logged.

        """

        super(ValueError, self).__init__("ValueError", error)



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="statys",
    version="1.0.1",
    description="Statistical Analyzers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Gustavo Rosa",
    author_email="gustavo.rosa@unesp.br",
    url="https://github.com/gugarosa/statys",
    license="Apache 2.0",
    install_requires=[
        "coverage>=5.5",
        "matplotlib>=3.3.4",
        "numpy>=1.19.5",
        "pre-commit>=2.17.0",
        "pylint>=2.7.2",
        "pytest>=6.2.2" "scipy>=1.5.4",
    ],
    extras_require={
        "tests": [
            "coverage",
            "pytest",
            "pytest-pep8",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(),
)



"""Friedman-related tests and their post-hocs.
"""

import numpy as np

import statys.tests.measure as m
import statys.utils.constants as c
from statys.utils import logging

logger = logging.get_logger(__name__)


def friedman(dist, **kwargs):
    """Performs the Friedman test.

    Args:
        dist (Distribution): Distribution to be analyzed.

    Returns:
        Dictionary holding the test's outputs.

    """

    logger.info("Performing Friedman test ...")

    # Initializes a empty dictionary for the outputs
    output = {}

    # Computes the average ranks (axis keyword should be used accordingly)
    average_ranks = m.rank(dist, **kwargs)

    for key, val in average_ranks.items():
        # Initializes the number of measurements as one
        n = 1

        # If argument has more than one measurement
        if val.ndim > 1:
            # Gathers the number of measurements
            n = val.shape[0]

            # Averages its rankings
            val = np.mean(val, axis=0)

        # Gathers the amount of ranks
        k = len(val)

        # Calculates the Friedman's statistic
        f = (
            12
            * n
            * (sum([v**2.0 for v in val]) - (k * (k + 1) * (k + 1) / 4))
            / (k * (k + 1))
        )

        # Calculates the F-distribution
        f_dist = (k - 1, (k - 1) * (n - 1))

        # Calculates the Iman's statistic
        iman = (n - 1) * f / (n * (k - 1) - f)

        # Adds the tuple to the output dictionary
        output[key] = (f, k - 1), (iman, f_dist)

    logger.info("Test performed.")
    logger.debug(output)

    return output


def friedman_with_posthoc(dist, alpha=0.05, post_hoc="nemenyi", **kwargs):
    """Performs the Friedman test with a post-hoc analysis.

    Args:
        dist (Distribution): Distribution to be analyzed.
        alpha (float): Significance value.
        post_hoc (str): Type of post-hoc analysis.

    Returns:
        Dictionary holding the test's outputs.

    """

    logger.info("Performing Friedman-%s test ...", post_hoc)

    # Initializes a empty dictionary for the outputs
    output = {}

    # Computes the average ranks (axis keyword should be used accordingly)
    average_ranks = m.rank(dist, **kwargs)

    if alpha == 0.01:
        critical_index = 0

    elif alpha == 0.05:
        critical_index = 1

    elif alpha == 0.1:
        critical_index = 2

    # Gathers the corresponding critical values
    q = np.asarray(c.CRITICAL_VALUES[post_hoc])[:, critical_index]

    for key, val in average_ranks.items():
        # Initializes the number of measurements as one
        n = 1

        # If argument has more than one measurement
        if val.ndim > 1:
            # Gathers the number of measurements
            n = val.shape[0]

            # Averages its rankings
            val = np.mean(val, axis=0)

        # Gathers the amount of ranks
        k = len(val)

        # Calculates the critical difference
        cd = q[k - 1] * (k * (k + 1) / (6 * n)) ** 0.5

        # Adds the tuple to the output dictionary
        output[key] = (val, cd)

    logger.info("Test performed.")
    logger.debug(output)

    return output
